@@ -205,7 +205,7 @@ headers and provide form."
 (ert-deftest package-lint-test-warn-lexical-binding-without-emacs-24-dep ()
   (should
    (equal
-    '((1 27 warning "You should depend on (emacs \"24\") if you need lexical-binding."))
+    '((1 27 warning "You should depend on (emacs \"24.1\") if you need lexical-binding."))
     (package-lint-test--run
      ""
      ";;; test.el --- A test -*- lexical-binding: t -*-\n"))))
@@ -269,9 +269,15 @@ headers and provide form."
     '((6 23 error "You can only depend on Emacs version 24 or greater: package.el for Emacs 23 does not support the \"emacs\" pseudopackage."))
     (package-lint-test--run ";; Package-Requires: ((emacs \"23\"))"))))
 
+(ert-deftest package-lint-test-warning-emacs-27-dep ()
+  (should
+   (equal
+    '((6 23 warning "This makes the package uninstallable in all released Emacs versions."))
+    (package-lint-test--run ";; Package-Requires: ((emacs \"27\"))"))))
+
 (ert-deftest package-lint-test-accept-emacs-24+-dep ()
   (should (equal '() (package-lint-test--run ";; Package-Requires: ((emacs \"24\"))")))
-  (should (equal '() (package-lint-test--run ";; Package-Requires: ((emacs \"26.7\"))"))))
+  (should (equal '() (package-lint-test--run ";; Package-Requires: ((emacs \"26.3\"))"))))
 
 (ert-deftest package-lint-test-error-uninstallable-dep ()
   (should
@@ -327,6 +333,16 @@ headers and provide form."
 Alternatively, depend on (emacs \"24.3\") or greater, in which cl-lib is bundled.")
     (package-lint-test--run ";; Package-Requires: ((cl-lib \"1\"))"))))
 
+(ert-deftest package-lint-test-warning-cl-lib-not-needed ()
+  (should
+   (member
+    '(6 52 warning "An explicit dependency on cl-lib <= 1.0 is not needed on Emacs >= 24.3.")
+    (package-lint-test--run ";; Package-Requires: ((emacs \"25.1\") (cl-lib \"1.0\"))")))
+  (should
+   (member
+    '(6 52 warning "An explicit dependency on cl-lib <= 1.0 is not needed on Emacs >= 24.3.")
+    (package-lint-test--run ";; Package-Requires: ((emacs \"24.3\") (cl-lib \"0.5\"))"))))
+
 (ert-deftest package-lint-test-warning-cl ()
   (should
    (member
@@ -361,6 +377,33 @@ Alternatively, depend on (emacs \"24.3\") or greater, in which cl-lib is bundled
     (package-lint-test--run
      "(defconst test-fn #'window-resize)"))))
 
+(ert-deftest package-lint-test-error-new-functions-as-arg ()
+  (should
+   (equal
+    '((6 10 error "You should depend on (emacs \"24.1\") if you need `window-resize'."))
+    (package-lint-test--run
+     "(funcall 'window-resize foo)")))
+  (should
+   (equal
+    '((6 11 error "You should depend on (emacs \"24.1\") if you need `window-resize'."))
+    (package-lint-test--run
+     "(funcall #'window-resize foo)")))
+  (should
+   (equal
+    '((6 8 error "You should depend on (emacs \"24.1\") if you need `window-resize'."))
+    (package-lint-test--run
+     "(apply 'window-resize foo)")))
+  (should
+   (equal
+    '((6 9 error "You should depend on (emacs \"24.1\") if you need `window-resize'."))
+    (package-lint-test--run
+     "(apply #'window-resize foo)")))
+  (should
+   (equal
+    '((6 13 error "You should depend on (emacs \"24.1\") if you need `window-resize'."))
+    (package-lint-test--run
+     "(advice-add 'window-resize foo)"))))
+
 (ert-deftest package-lint-test-accept-new-functions-with-dep ()
   (should
    (equal
@@ -433,6 +476,15 @@ Alternatively, depend on (emacs \"24.3\") or greater, in which cl-lib is bundled
   (should (equal '() (package-lint-test--run
                       "(defun org-dblock-write:test ())"))))
 
+(ert-deftest package-lint-test-error-redefining-builtins ()
+  (should
+   (equal
+    '((8 0 error "Define compatibility functions with a prefix, e.g. \"test--setq-local\", and use `defalias' where they exist."))
+    (package-lint-test--run "(eval-when-compile
+  (unless (fboundp 'setq-local)
+    (defmacro setq-local (var val)
+      (list 'set (list 'make-local-variable (list 'quote var)) val))))"))))
+
 (ert-deftest package-lint-test-error-new-libraries ()
   (should
    (equal
@@ -451,6 +503,23 @@ Alternatively, depend on (emacs \"24.3\") or greater, in which cl-lib is bundled
     '((6 1 error "`spell-buffer' was removed in Emacs version 26.1."))
     (package-lint-test--run "(spell-buffer)"))))
 
+(ert-deftest package-lint-test-allow-removed-and-readded-functions ()
+  (should
+   (equal
+    ;; This is clunky, but won't currently fix
+    '((6 1 error "You should depend on (emacs \"24.1\") if you need `mm-url-encode-multipart-form-data'.")
+      (6 1 error "You should depend on (emacs \"25.1\") if you need `mm-url-encode-multipart-form-data'.")
+      (6 1 error "`mm-url-encode-multipart-form-data' was removed in Emacs version 24.3."))
+    (package-lint-test--run "(mm-url-encode-multipart-form-data)")))
+  (equal
+   ;; This is clunky, but won't currently fix
+   '((6 1 error "You should depend on (emacs \"25.1\") if you need `mm-url-encode-multipart-form-data'.")
+     (6 1 error "`mm-url-encode-multipart-form-data' was removed in Emacs version 24.3."))
+   (package-lint-test--run ";; Package-Requires: ((emacs \"24.1\"))\n(mm-url-encode-multipart-form-data)"))  (should
+   (equal
+    '()
+    (package-lint-test--run ";; Package-Requires: ((emacs \"25.1\"))\n(mm-url-encode-multipart-form-data)"))))
+
 (ert-deftest package-lint-test-accept-new-libraries-with-dep ()
   (should (equal '() (package-lint-test--run
                       ";; Package-Requires: ((emacs \"24.4\"))\n(require 'subr-x)"))))
@@ -669,5 +738,17 @@ Alternatively, depend on (emacs \"24.3\") or greater, in which cl-lib is bundled
     '()
     (package-lint-test--run "(hello 'world\n ;; a comment\n)"))))
 
+(ert-deftest package-lint-test-accept-valid-prefix-mappings ()
+  (should
+   (equal
+    '()
+    (package-lint-test--run "(defun org-foobar-test ()) (provide 'ox-foobar)" nil nil nil "ox-foobar" nil nil "ox-foobar"))))
+
+(ert-deftest package-lint-test-error-invalid-prefix-mappings ()
+  (should
+   (equal
+    '((6 0 error "\"org-foobaz-test\" doesn't start with package's prefix \"ox-foobar\"."))
+    (package-lint-test--run "(defun org-foobaz-test ()) (provide 'ox-foobar)" nil nil nil "ox-foobar" nil nil "ox-foobar"))))
+
 (provide 'package-lint-test)
 ;;; package-lint-test.el ends here
